#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-

# 2004-2006 merlin@ghostwheel.de

# Entpacken eines chk/vif Paketes geht z.B. mit "viftool split deuf.vif"
# das packt die oggs auf dem data02.chk der deutschen Frauenstimme aus.

# Andere Sprachen lassen sich nicht ganz auspacken da dort scheinbar
# andere Blocktypen (gepackt?) verwendet werden.

# Einpacken geht mit "viftool join 2 "Deutsch - Lisa" pfad/deuf.vif"
# das erzeugt die angegebene vif-datei mit der Beschreibung
# "Deutsch - Lisa" (das TomTom ignoriert die aber) und der chk Datei
# Nr. 2, also data02.chk (hier kannst du scheinbar nur dem TomTom
# bekannte Nummern verwenden, d.h. die die auf deiner TomTom-CD sind).

# Die Sprachdateien mssen alle 59 vorhanden sein und z.b.
# "pfad/RbLeft.ogg" heien und im Pfad liegen wo die vif datei erstellt
# werden soll. Die OGGs sollten mit 22050 Hz 1 Channel (mit z.B.
# "oggenc -b 40 bla.wav" erstellt) kodiert sein.

import sys, os, struct

filedescs = {
	 0: ('After',       'Nach'),
	 1: ('50',          '50'),
	 2: ('80',          '80'),
	 3: ('100',         '100'),
	 4: ('200',         '200'),
	 5: ('300',         '300'),
	 6: ('400',         '400'),
	 7: ('500',         '500'),
	 8: ('600',         '600'),
	 9: ('700',         '700'),
	10: ('800',         '800'),
	11: ('Meters',      'Metern'),
	12: ('Arrive',      'Sie haben ihren Bestimmungsort erreicht'),
	13: ('TurnRight',   'Biegen Sie rechts ab'),
	14: ('2ndLeft',     'Biegen Sie die zweite links ab'),
	15: ('2ndRight',    'Biegen Sie die zweite rechts ab'),
	16: ('3rdLeft',     'Biegen Sie die dritte links ab'),
	17: ('3rdRight',    'Biegen Sie die dritte rechts ab'),
	18: ('AhExit',      'Ausfahrt vor Ihnen'),
	19: ('AhExitLeft',  'Ausfahrt links vor Ihnen'),
	20: ('AhExitRight', 'Ausfahrt rechts vor Ihnen'),
	21: ('AhFerry',     'Nehmen Sie die Fhre vor Ihnen'),
	22: ('AhKeepLeft',  'Bleiben Sie links'),
	23: ('AhKeepRight', 'Bleiben Sie rechts'),
	24: ('AhLeftTurn',  'Abbiegung links vor Ihnen'),
	25: ('AhRightTurn', 'Abbiegung rechts vor Ihnen'),
	26: ('AhUTurn',     'Wendemglichkeit vor Ihnen'),
	27: ('BearLeft',    'Halten Sie sich links'),
	28: ('BearRight',   'Halten Sie sich rechts'),
	29: ('Charge',      'Congestion charge'),
	30: ('Depart',      'Abfahrt'),
	31: ('KeepLeft',    'Bleiben Sie links'),
	32: ('KeepRight',   'Bleiben Sie rechts'),
	33: ('LnLeft',      'Bleiben Sie auf der linken Spur'),
	34: ('LnRight',     'Bleiben Sie auf der rechten Spur'),
	35: ('MwEnter',     'Fahren Sie auf die Autobahn'),
	36: ('MwExit',      'Nehmen Sie die Ausfahrt'),
	37: ('MwExitLeft',  'Nehmen Sie die Ausfahrt links'),
	38: ('MwExitRight', 'Nehmen Sie die Ausfahrt rechts'),
	39: ('RbBack',      'Fahren Sie um den Kreisverkehr herum'),
	40: ('RbCross',     'Fahren Sie geradeaus ber den Kreisverkehr'),
	41: ('RbExit1',     'Erste Ausfahrt'),
	42: ('RbExit2',     'Zweite Ausfahrt'),
	43: ('RbExit3',     'Dritte Ausfahrt'),
	44: ('RbExit4',     'Vierte Ausfahrt'),
	45: ('RbExit5',     'Fnfte Ausfahrt'),
	46: ('RbExit6',     'Sechste Ausfahrt'),
	47: ('RbLeft',      'Biegen Sie im Kreisverkehr links ab'),
	48: ('RbRight',     'Biegen Sie im Kreisverkehr rechts ab'),
	49: ('RoadEnd',     'Am Ende der Strae'),
	50: ('SharpLeft',   'scharf links'),
	51: ('SharpRight',  'scharf rechts'),
	52: ('Straight',    'Fahren Sie geradeaus'),
	53: ('TakeFerry',   'Fahren Sie mit der Fhre'),
	54: ('Then',        'dann'),
	55: ('TryUTurn',    'Drehen Sie wenn mglich um'),
	56: ('TurnLeft',    'Biegen Sie links ab'),
	57: ('UTurn',       'Wenden Sie'),
	58: ('Yards',       'Yards'),
}

silent = False
def write(*texts):
	if not silent:
		sys.stdout.write(' '.join(map(str,texts)))
		sys.stdout.write('\n')
		sys.stdout.flush()

def vifchk(file):
	fext = os.path.splitext(file)[1].lower()
	path = os.path.split(file)[0]
	if fext == '.vif':
		try:
			fdesc, datafile = filter(None, map(lambda l:l.strip(), open(file,'r').readlines()))[:2]
		except Exception, reason:
			write('Invalid vif file: %s!'%(file))
			write('[%s]'%(str(reason)))
			return path, None
		else:
			fdesc = fdesc.strip()
			datafile = os.path.join(path, datafile.strip())
			write('DESC: %s'%(fdesc))
			return path, datafile
	elif fext == '.chk':
		return path, file
	else:
		write('Invalid file: %s!'%(file))
		return path, None

def split(file, intern=None):

	res = []
	rem = {}
	path, datafile = vifchk(file)

	if datafile is None:
		return res

	write('DATA: %s'%(datafile))

	flen = os.stat(datafile)[6]
	inp = open(datafile,'rb')

	# read header
	num = struct.unpack('>L',inp.read(4))[0]
	entries=[]
	write('Found %d entries...'%(num))
	if (num+1)*4 > flen:
		write('Invalid chk file!')
		return

	while num > 0:
		entries.append(struct.unpack('>L',inp.read(4))[0])
		num -= 1

	# check length
	fend = struct.unpack('>L',inp.read(4))[0]
	if fend != flen:
		write('File length mismatch (is %d, should be %d).'%(flen, fend))
		return

	write('Extracting entries...')

	# check entries
	cnt=0
	for pos in entries:
		if cnt+1 < len(entries):
			next = entries[cnt+1]
		else:
			next = flen
		inp.seek(pos,0)
		tag, typ, blocks = struct.unpack('>BBH', inp.read(4))
		err=[]
		# v debug
		if typ == 4:
			head1, head2, head3, length = struct.unpack('>LLBL', inp.read(13))
			write('%4d'%cnt, '@ %08x :'%pos, '%02x'%tag, '%02x'%typ, '%04x'%blocks, '%08x'%head1, '%08x'%head2, '%02x'%head3, ':', '%6d'%length, '=', '0x%x'%length, ': blocks:', '%6d'%(blocks*4), 'space:', next-pos)
		else:
			head1, head2, length = struct.unpack('>LLL', inp.read(12))
			write('%4d'%cnt, '@ %08x :'%pos, '%02x'%tag, '%02x'%typ, '%04x'%blocks, '%08x'%head1, '%08x'%head2, '-- :', '%6d'%length, '=', '0x%x'%length, ': blocks:', '%6d'%(blocks*4), 'space:', next-pos)
		# ^ debug
		if (next-pos)%4 != 0:
			err.append('space is not a multiple of 4')
		if length > next-pos:
			err.append('length is bigger than space')
		if blocks*4 > next-pos:
			err.append('blocks is bigger than space')
		if blocks*4 != next-pos-4:
			err.append('blocks should be space-4')
		if err:
			write('!!!', '\n!!! '.join(err))
		elif filedescs.has_key(cnt):
			data = inp.read(length)
			if intern is None:
				open(os.path.join(path, '%s.ogg'%filedescs[cnt][0]), 'wb').write(data)
			else:
				if cnt in intern:
					rem[cnt] = data
		else:
			write('unknown voice file #%d'%(cnt))
		cnt+=1
	inp.close()
	if intern is not None:
		for number in intern:
			res.append((number, rem[number]))
	return res

def join(chknum, fdesc, file):
	path = os.path.split(file)[0]
	if chknum is None:
		chkfile = file
	else:
		chkfile = os.path.join(path, 'data%02d.chk'%(int(chknum)))
	write('Gathering sound files...')
	entries = []
	for number, (name, desc) in filedescs.items():
		sndfile = os.path.join(path,name)
		if os.path.isfile(name+'.ogg'):
			entries.append(open(sndfile+'.ogg','rb').read())
		elif os.path.isfile(name+'.wav'):
			entries.append(os.popen('oggenc -b 40 -o - %s.wav 2>/dev/null'%(sndfile),'rb').read())
		else:
			write('entry #%d "%s.ogg" (%s) is missing'%(number,name,desc))
			return
	if chknum is not None:
		write('Creating vif file...')
		open(file,'wb').write('%s\r\n%s\r\n'%(fdesc,os.path.split(chkfile)[1]))
	write('Creating chk file...')
	out = open(chkfile,'wb')
	# write entry-count
	out.write(struct.pack('>L', len(filedescs)))
	# set and write entry start positions
	pos = 8+len(entries)*4
	for entry in entries:
		out.write(struct.pack('>L', pos))
		elen = len(entry)
		pos += 16 + elen+3-((elen-1)%4)
	# write file end marker
	out.write(struct.pack('>L', pos))
	# write entry data
	for entry in entries:
		elen = len(entry)
		pad = 3-(elen-1)%4
		out.write(struct.pack('>BBHLLL', 1, 0, (elen+12+pad)>>2, 1, 8, elen))
		out.write(entry+'\000'*pad)
	out.close()
	write('Done.')

def speak(file, *text):

	try:
		numbers = map(int, text)
	except:
		write('Invalid text - only numbers supported yet!')
		return

	global silent
	silent = True
	samples = split(file, numbers)
	silent = False

	if samples:
		command = ['ogg123 -q']
		for number, sample in samples:
			file = '%d.speak.tmp.ogg'%(number)
			open(file,'wb').write(sample)
			command.append(file)
		os.system(' '.join(command))
		for number, sample in samples:
			os.unlink('%d.speak.tmp.ogg'%(number))

## MAIN ##

if len(sys.argv)<2 or sys.argv[1] not in ['split', 'join', 'speak'] or (sys.argv[1]=='join' and len(sys.argv) not in [3,5]):
	myself = os.path.split(sys.argv[0])[1]
	write('Usage: %s split [VIFFILE|CHKFILE]+'%(myself))
	write('       %s join [NUM DESC VIFFILE|CHKFILE]'%(myself))
	write('       %s speak [VIFFILE|CHKFILE] [NUMBER|TEXT]+'%(myself))
	sys.exit()
else:
	action = sys.argv[1]

if action == 'split':
	for file in sys.argv[2:]:
		split(file)

elif action == 'join':
	if len(sys.argv) == 5:
		join(*sys.argv[2:5])
	else:
		join(None, None, sys.argv[2])

elif action == 'speak':
	speak(*sys.argv[2:])

